# Copyright (C) 2013 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of pytest-dbfixtures.

# pytest-dbfixtures is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pytest-dbfixtures is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with pytest-dbfixtures.  If not, see <http://www.gnu.org/licenses/>.

import os
import pytest
import subprocess

from path import path
from tempfile import mkdtemp

from pytest_dbfixtures import factories
from pytest_dbfixtures.executors import TCPExecutor
from pytest_dbfixtures.utils import get_config, try_import


ROOT_DIR = path(__file__).parent.parent.abspath()


def pytest_addoption(parser):
    parser.addoption(
        '--dbfixtures-config',
        action='store',
        default=ROOT_DIR / 'pytest_dbfixtures' / 'conf' / 'dbfixtures.conf',
        metavar='path',
        dest='db_conf',
    )

    parser.addoption(
        '--mongo-config',
        action='store',
        default=ROOT_DIR / 'pytest_dbfixtures' / 'conf' / 'mongo.conf',
        metavar='path',
        dest='mongo_conf',
    )

    parser.addoption(
        '--redis-config',
        action='store',
        default=ROOT_DIR / 'pytest_dbfixtures' / 'conf' / 'redis.conf',
        metavar='path',
        dest='redis_conf',
    )

    parser.addoption(
        '--rabbit-config',
        action='store',
        default=ROOT_DIR / 'pytest_dbfixtures' / 'conf' / 'rabbit.conf',
        metavar='path',
        dest='rabbit_conf',
    )


redis_proc = factories.redis_proc()
redisdb = factories.redisdb('redis_proc')

postgresql_proc = factories.postgresql_proc()
postgresql = factories.postgresql('postgresql_proc')

mysql_proc = factories.mysql_proc()
mysqldb = factories.mysqldb('mysql_proc', scope='function')
mysqldb_session = factories.mysqldb('mysql_proc')

elasticsearch_proc = factories.elasticsearch_proc()
elasticsearch = factories.elasticsearch('elasticsearch_proc')


@pytest.fixture(scope='session')
def mongo_proc(request):
    """
    #. Get config.
    #. Run a ``mongod`` process.
    #. Stop ``mongod`` process after tests.

    .. note::
        `mongod <http://docs.mongodb.org/v2.2/reference/mongod/>`_

    :param FixtureRequest request: fixture request object
    :rtype: pytest_dbfixtures.executors.TCPExecutor
    :returns: tcp executor
    """
    config = get_config(request)
    mongo_conf = request.config.getvalue('mongo_conf')

    mongo_executor = TCPExecutor(
        '{mongo_exec} {params} {config}'.format(
            mongo_exec=config.mongo.mongo_exec,
            params=config.mongo.params,
            config=mongo_conf),
        host=config.mongo.host,
        port=config.mongo.port,
    )
    mongo_executor.start()

    def stop():
        mongo_executor.stop()

    request.addfinalizer(stop)

    return mongo_executor


@pytest.fixture
def mongodb(request, mongo_proc):
    """
    #. Get pymongo module and config.
    #. Get connection to mongo.
    #. Drop collections before and after tests.

    :param FixtureRequest request: fixture request object
    :param TCPExecutor mongo_proc: tcp executor
    :rtype: pymongo.connection.Connection
    :returns: connection to mongo database
    """
    pymongo, config = try_import('pymongo', request)

    mongo_conn = pymongo.Connection(
        config.mongo.host,
        config.mongo.port
    )

    def drop():
        for db in mongo_conn.database_names():
            for collection_name in mongo_conn[db].collection_names():
                if collection_name != 'system.indexes':
                    mongo_conn[db][collection_name].drop()

    request.addfinalizer(drop)

    drop()

    return mongo_conn


def get_rabbit_env(name):
    """
    Get value from environment variable. If does not exists (older version)
    then use older name.

    :param str name: name of environment variable
    :rtype: str
    :returns: path to directory
    """
    return os.environ.get(name) or os.environ.get(name.split('RABBITMQ_')[1])


def get_rabbit_path(name):
    """
    Get a path to directory contains sub-directories for the RabbitMQ
    server's Mnesia database files.
    `Relocate <http://www.rabbitmq.com/relocate.html>`_

    If environment variable or path to directory do not exist, return ``None``,
    else return path to directory.

    :param str name: name of environment variable
    :rtype: path.path or None
    :returns: path to directory
    """
    env = get_rabbit_env(name)

    if not env or not path(env).exists():
        return

    return path(env)


@pytest.fixture
def rabbitmq_proc(request):
    """
    #. Get config.
    #. Make a temporary directory.
    #. Setup environment variables.
    #. Start a rabbit server
        `<http://www.rabbitmq.com/man/rabbitmq-server.1.man.html>`_
    #. Stop rabbit server and remove temporary files after tests.

    :param FixtureRequest request: fixture request object
    :rtype: pytest_dbfixtures.executors.TCPExecutor
    :returns: tcp executor
    """

    rabbit_conf = open(request.config.getvalue('rabbit_conf')).readlines()
    rabbit_conf = dict(line[:-1].split('=') for line in rabbit_conf)

    tmpdir = path(mkdtemp(prefix='rabbitmq_fixture'))
    rabbit_conf['RABBITMQ_LOG_BASE'] = str(tmpdir)
    rabbit_conf['RABBITMQ_MNESIA_BASE'] = str(tmpdir)

    for name, value in rabbit_conf.items():
        # for new versions of rabbitmq-server
        os.environ[name] = value
        # for older versions of rabbitmq-server
        prefix, name = name.split('RABBITMQ_')
        os.environ[name] = value

    pika, config = try_import('pika', request)

    rabbit_executor = TCPExecutor(
        config.rabbit.rabbit_server,
        config.rabbit.host,
        config.rabbit.port,
    )
    base_path = get_rabbit_path('RABBITMQ_MNESIA_BASE')

    def stop_and_reset():
        rabbit_executor.stop()
        base_path.rmtree()
        tmpdir.exists() and tmpdir.rmtree()

    request.addfinalizer(stop_and_reset)

    rabbit_executor.start()
    pid_file = base_path / get_rabbit_env('RABBITMQ_NODENAME') + '.pid'
    wait_cmd = config.rabbit.rabbit_ctl, '-q', 'wait', pid_file
    subprocess.Popen(wait_cmd).communicate()

    return rabbit_executor


@pytest.fixture
def rabbitmq(rabbitmq_proc, request):
    """
    #. Get module and config.
    #. Connect to RabbitMQ using the parameters from config.

    :param TCPExecutor rabbitmq_proc: tcp executor
    :param FixtureRequest request: fixture request object
    :rtype: pika.adapters.blocking_connection.BlockingConnection
    :returns: instance of :class:`BlockingConnection`
    """
    pika, config = try_import('pika', request)

    rabbit_params = pika.connection.ConnectionParameters(
        host=config.rabbit.host,
        port=config.rabbit.port,
        connection_attempts=3,
        retry_delay=2,
    )

    try:
        rabbit_connection = pika.BlockingConnection(rabbit_params)
    except pika.adapters.blocking_connection.exceptions.ConnectionClosed:
        print "Be sure that you're connecting rabbitmq-server >= 2.8.4"

    return rabbit_connection



# Copyright (C) 2013 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of pytest-dbfixtures.

# pytest-dbfixtures is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pytest-dbfixtures is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with pytest-dbfixtures.  If not, see <http://www.gnu.org/licenses/>.

import os
import re
import shutil
import subprocess
import time

import pytest

from pytest_dbfixtures.executors import TCPExecutor
from pytest_dbfixtures.utils import get_config, try_import


START_INFO = 'database system is ready to accept connections'

BASE_PROC_START_COMMAND = """
{postgresql_ctl} start -D {datadir}
-o "-F -p {port} -c %s='{unixsocketdir}'"
-l {logfile} {startparams}"""

PROC_START_COMMAND = {
    '8.4': BASE_PROC_START_COMMAND % 'unix_socket_directory',
    '9.0': BASE_PROC_START_COMMAND % 'unix_socket_directory',
    '9.1': BASE_PROC_START_COMMAND % 'unix_socket_directory',
    '9.2': BASE_PROC_START_COMMAND % 'unix_socket_directory',
    '9.3': BASE_PROC_START_COMMAND % 'unix_socket_directories',
}


def wait_for_postgres(config, awaited_msg):
    """
    Blocks until logfile is created by psql server.
    Awaits for given message in logfile. Block until log contains message.

    :param pymlconf.ConfigManager config: config
    :param str awaited_msg: awaited message
    """
    while not os.path.isfile(config.postgresql.logfile):
        time.sleep(1)

    while 1:
        with open(config.postgresql.logfile, 'r') as content_file:
            content = content_file.read()
            if awaited_msg in content:
                break
        time.sleep(1)


def remove_postgresql_directory(config):
    """
    Checks postgresql directory and logfile. Delete a logfile if exist.
    Recursively delete a directory tree if exist.

    :param pymlconf.ConfigManager config: config
    """
    if os.path.isfile(config.postgresql.logfile):
        os.remove(config.postgresql.logfile)
    if os.path.isdir(config.postgresql.datadir):
        shutil.rmtree(config.postgresql.datadir)


def init_postgresql_directory(config):
    """
    #. Remove postgresql directory if exist.
    #. `Initialize postgresql data directory
        <www.postgresql.org/docs/9.1/static/app-initdb.html>`_

    :param pymlconf.ConfigManager config: config

    """
    remove_postgresql_directory(config)
    init_directory = (
        config.postgresql.postgresql_ctl, 'initdb',
        '-o "--auth=trust --username=%s"' % config.postgresql.user,
        '-D %s' % config.postgresql.datadir,
    )
    subprocess.check_output(' '.join(init_directory), shell=True)


def init_postgresql_database(postgresql, config):
    """
    #. Connect to psql with proper isolation level
    #. Create test database
    #. Close connection

    :param FixtureRequest postgresql: psycopg2 object
    :param pymlconf.ConfigManager config: config

    """

    conn = postgresql.connect(
        user=config.postgresql.user,
        host=config.postgresql.host,
        port=config.postgresql.port
    )
    conn.set_isolation_level(postgresql.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    cur = conn.cursor()
    cur.execute('CREATE DATABASE ' + config.postgresql.db)
    cur.close()
    conn.close()


def drop_postgresql_database(postgresql, config):
    """
    #. Connect to psql with proper isolation level
    #. Drop test database
    #. Close connection

    :param FixtureRequest postgresql: psycopg2 object
    :param pymlconf.ConfigManager config: config
    """
    conn = postgresql.connect(
        user=config.postgresql.user,
        host=config.postgresql.host,
        port=config.postgresql.port
    )
    conn.set_isolation_level(postgresql.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
    cur = conn.cursor()
    cur.execute('DROP DATABASE IF EXISTS %s' % config.postgresql.db)
    cur.close()
    conn.close()


def postgresql_version(pg_ctl):
    """
    Detect postgresql version.

    :return: minor version string
    :rtype: str
    """
    version_string = subprocess.check_output([pg_ctl, '--version'])
    matches = re.search('.* (?P<version>\d\.\d)\.\d', version_string)
    return matches.groupdict()['version']


def postgresql_proc(executable=None, host=None, port=None):
    """
    postgresql process factory.

    :param str executable: path to postgresql_ctl
    :param str host: hostname
    :param str port: port
    :rtype: func
    :returns: function which makes a postgresql process
    """

    @pytest.fixture(scope='session')
    def postgresql_proc_fixture(request):
        """
        #. Get config.
        #. Initialize postgresql data directory
        #. `Start a postgresqld server
            <http://www.postgresql.org/docs/9.1/static/app-pg-ctl.html>`_
        #. Stop server and remove directory after tests.
            `See <http://www.postgresql.org/docs/9.1/static/app-pg-ctl.html>`_

        :param FixtureRequest request: fixture request object
        :rtype: pytest_dbfixtures.executors.TCPExecutor
        :returns: tcp executor
        """
        config = get_config(request)
        config.postgresql.postgresql_ctl = executable or config.postgresql.postgresql_ctl  # noqa
        # check if that executable exists, as it's no on system PATH
        if not os.path.exists(config.postgresql.postgresql_ctl) and (
            # only replace if executable isn't passed manually
            executable is None
        ):
            pg_bindir = subprocess.check_output(
                ['pg_config', '--bindir'], universal_newlines=True).strip()
            config.postgresql.postgresql_ctl = os.path.join(
                pg_bindir, 'pg_ctl')

        config.postgresql.host = host or config.postgresql.host
        config.postgresql.port = port or config.postgresql.port
        config.postgresql.datadir += str(config.postgresql.port)
        config.postgresql.logfile = '/tmp/postgresql.{0}.log'.format(
            config.postgresql.port
        )

        init_postgresql_directory(config)
        pg_version = postgresql_version(config.postgresql.postgresql_ctl)
        postgresql_executor = TCPExecutor(
            PROC_START_COMMAND[pg_version].format(
                postgresql_ctl=config.postgresql.postgresql_ctl,
                datadir=config.postgresql.datadir,
                port=config.postgresql.port,
                unixsocketdir=config.postgresql.unixsocketdir,
                logfile=config.postgresql.logfile,
                startparams=config.postgresql.startparams,
            ),
            host=config.postgresql.host,
            port=config.postgresql.port,
        )
        postgresql_executor.start()
        if '-w' in config.postgresql.startparams:
            wait_for_postgres(config, START_INFO)

        def shutdown_server():
            return (
                config.postgresql.postgresql_ctl, 'stop',
                '-D %s' % config.postgresql.datadir,
                '-o "-p %s -c unix_socket_directory=\'%s\'"' % (
                    config.postgresql.port, config.postgresql.unixsocketdir)
            )

        def stop_server_and_remove_directory():
            subprocess.check_output(' '.join(shutdown_server()), shell=True)
            postgresql_executor.stop()
            remove_postgresql_directory(config)

        request.addfinalizer(stop_server_and_remove_directory)

        return postgresql_executor

    return postgresql_proc_fixture


def postgresql(process_fixture_name, host=None, port=None, db=None):
    """
    postgresql database factory.

    :param str process_fixture_name: name of the process fixture
    :param str host: hostname
    :param int port: port
    :param int db: database name
    :rtype: func
    :returns: function which makes a connection to postgresql
    """

    @pytest.fixture
    def postgresql_factory(request):
        """
        #. Load required process fixture.
        #. Get postgresql module and config.
        #. Connect to postgresql.
        #. Flush database after tests.

        :param FixtureRequest request: fixture request object
        :rtype: psycopg2.connection
        :returns: postgresql client
        """
        request.getfuncargvalue(process_fixture_name)

        postgresql, config = try_import('psycopg2', request)
        config.postgresql.host = host or config.postgresql.unixsocketdir
        config.postgresql.port = port or config.postgresql.port
        config.postgresql.db = db or config.postgresql.db

        init_postgresql_database(postgresql, config)
        conn = postgresql.connect(
            dbname=config.postgresql.db,
            user=config.postgresql.user,
            host=config.postgresql.host,
            port=config.postgresql.port
        )

        def drop_database():
            conn.close()
            drop_postgresql_database(postgresql, config)

        request.addfinalizer(drop_database)

        return conn

    return postgresql_factory


__all__ = [postgresql, postgresql_proc]



# Copyright (C) 2013-2014 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of pytest-dbfixtures.

# pytest-dbfixtures is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pytest-dbfixtures is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with pytest-dbfixtures.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from pytest_dbfixtures.executors import HTTPExecutor
from pytest_dbfixtures.utils import get_config, try_import


def elasticsearch_proc(host='127.0.0.1', port=9201, cluster_name=None):
    """
    Creates elasticsearch process fixture.

    .. warning::

        This fixture requires at least version 1.0 of elasticsearch to work.

    :param str host: host that the instance listens on
    :param int port: port that the instance listens on
    :param str cluster_name: name of a cluser this node should work on.
        Used for autodiscovery. By default each node is in it's own cluser.
    """
    @pytest.fixture(scope='session')
    def elasticsearch_proc_fixture(request):
        """
        Elasticsearch process starting fixture.
        """
        config = get_config(request)

        pidfile = '/tmp/elasticsearch.{0}.pid'.format(port)
        home_path = '/tmp/elasticsearch_{0}'.format(port)
        logs_path = '/tmp/elasticsearch_{0}_logs'.format(port)
        work_path = '/tmp/elasticsearch_{0}_tmp'.format(port)
        cluster = cluster_name or 'dbfixtures.{0}'.format(port)

        command_exec = '''
            {deamon} -p {pidfile} -Des.http.port={port}
            -Des.path.home={home_path}  -Des.default.path.logs={logs_path}
            -Des.default.path.work={work_path}
            -Des.default.path.conf=/etc/elasticsearch
            -Des.cluster.name={cluster}
            '''.format(
            deamon=config.elasticsearch.deamon,
            pidfile=pidfile,
            port=port,
            home_path=home_path,
            logs_path=logs_path,
            work_path=work_path,
            cluster=cluster
        )

        elasticsearch_executor = HTTPExecutor(
            command_exec, 'http://{host}:{port}'.format(
                host=host,
                port=port
            ),
        )

        elasticsearch_executor.start()

        request.addfinalizer(elasticsearch_executor.stop)
        return elasticsearch_executor

    return elasticsearch_proc_fixture


def elasticsearch(proc_fixture, hosts='127.0.01:9201'):
    """
    Creates Elasticsearch client fixture.

    :param hosts: elasticsearch hosts list. See
        http://elasticsearch-py.readthedocs.org/en/master/api.html#elasticsearch.Elasticsearch for details. # noqa
    """

    @pytest.fixture
    def elasticsearch_fixture(request):
        """Elasticsearch client fixture."""
        request.getfuncargvalue(proc_fixture)

        elasticsearch, _ = try_import('elasticsearch', request)

        client = elasticsearch.Elasticsearch(hosts=hosts)

        def drop_indexes():
            client.indices.delete(index='*')

        request.addfinalizer(drop_indexes)

        return client

    return elasticsearch_fixture



# Copyright (C) 2013 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of pytest-dbfixtures.

# pytest-dbfixtures is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pytest-dbfixtures is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with pytest-dbfixtures.  If not, see <http://www.gnu.org/licenses/>.

'''
Overrides summon_process executors to defaultly use wait=True parameter
which turns on waiting on process beeing actually closed and set default
timeout for process to start.
'''

from summon_process.executors import (
    SimpleExecutor,
    TCPCoordinatedExecutor,
    HTTPCoordinatedExecutor,
)


DEFAULT_TIMEOUT = 100


class BaseExecutor(SimpleExecutor):

    '''
        Set timeout=DEFAULT_TIMEOUT on __init__ and wait=True on stop()
        on executors inheriting from this class.
    '''

    def __init__(self, *args, **kwargs):
        if kwargs.get('timeout') is None:
            kwargs['timeout'] = DEFAULT_TIMEOUT
        super(BaseExecutor, self).__init__(*args, **kwargs)

    def stop(self, *args, **kwargs):
        if 'wait' not in kwargs:
            kwargs['wait'] = True
        super(BaseExecutor, self).stop(*args, **kwargs)


class TCPExecutor(BaseExecutor, TCPCoordinatedExecutor):
    pass


class HTTPExecutor(BaseExecutor, HTTPCoordinatedExecutor):
    pass



# Copyright (C) 2013 by Clearcode <http://clearcode.cc>
# and associates (see AUTHORS).

# This file is part of pytest-dbfixtures.

# pytest-dbfixtures is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# pytest-dbfixtures is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with pytest-dbfixtures.  If not, see <http://www.gnu.org/licenses/>.

import re
import os
from setuptools import setup, find_packages


here = os.path.dirname(__file__)
with open(os.path.join(here, 'pytest_dbfixtures', '__init__.py')) as v_file:
    package_version = re.compile(
        r".*__version__ = '(.*?)'", re.S).match(v_file.read()).group(1)


setup(
    name='pytest-dbfixtures',
    version=package_version,
    description='Databases fixtures plugin for py.test.',
    author='Clearcode - The A Room',
    author_email='thearoom@clearcode.cc',
    url='https://github.com/clearcode/pytest-dbfixtures',
    packages=find_packages(),
    install_requires=[
        'pytest>=2.3.4',
        'summon_process>=0.1.3',
        'pyaml>=3.10',
        'pymlconf>=0.2.10a',
        'path.py>=4.2',
    ],
    include_package_data=True,
    extras_require={
        'documentation': [
            "Sphinx",
        ]
    },
    entry_points={
        'pytest11': [
            'pytest_dbfixtures = pytest_dbfixtures.plugin'
        ]},
    keywords='py.test pytest fixture redis mongo rabbitmq mysql postgresql',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development :: Testing',
        'Topic :: Utilities',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7']
)
